
################################################################################
#
# Orthogonal direct sums and embeddings of orthogonal groups
#
################################################################################

# this function is used whenever A and B are in direct orthogonal sum in a
# bigger torsion module. Let D be this sum. Since A and B are in orthogonal
# direct sum in D, we can embed O(A) and O(B) in O(D).
#
# This function returns D, the embeddings A\to D and B\to D, as well as O(D)
# together with the embeddings O(A) \to O(D) and O(B) \to O(D)
function _sum_with_embeddings_orthogonal_groups(A::TorQuadModule, B::TorQuadModule)
  D = A+B
  AinD = hom(A, D, TorQuadModuleElem[D(lift(a)) for a in gens(A)])
  BinD = hom(B, D, TorQuadModuleElem[D(lift(b)) for b in gens(B)])
  @assert all(AinD(a)*BinD(b) == 0 for a in gens(A), b in gens(B))
  OD = orthogonal_group(D)
  OA = orthogonal_group(A)
  OB = orthogonal_group(B)

  gene = data.(union!(AinD.(gens(A)), BinD.(gens(B))))
  geneOAinOD = elem_type(OD)[]
  for f in gens(OA)
    imgf = data.(union!(AinD.(f.(gens(A))), BinD.(gens(B))))
    fab = hom(gene, imgf)
    fD = OD(hom(D, D, fab.map))
    push!(geneOAinOD, fD)
  end

  geneOBinOD = elem_type(OD)[]
  for f in gens(OB)
    imgf = data.(union!(AinD.(gens(A)), BinD.(f.(gens(B)))))
    fab = hom(gene, imgf)
    fD = OD(hom(D, D, fab.map))
    push!(geneOBinOD, fD)
  end
  OAtoOD = hom(OA, OD, geneOAinOD; check = false)
  OBtoOD = hom(OB, OD, geneOBinOD; check = false)
  return D, AinD, BinD, OD, OAtoOD, OBtoOD
end

# Construct the direct sum D of A and B. Since the images of A and B are in
# orthogonal direct sum, we can embed O(A) in O(D) and O(B) in O(D).
#
# This function returns D together with the injections A \to D and B \to D, as
# well as O(D) with the embeddings O(A) \to O(D) and O(B) \to O(D).
function _direct_sum_with_embeddings_orthogonal_groups(A::TorQuadModule, B::TorQuadModule)
  D, inj = direct_sum(A, B)
  AinD, BinD = inj
  OD = orthogonal_group(D)
  OA = orthogonal_group(A)
  OB = orthogonal_group(B)

  geneOAinOD = elem_type(OD)[]
  for f in gens(OA)
    m = block_diagonal_matrix(ZZMatrix[matrix(f), identity_matrix(ZZ, ngens(B))])
    fD = OD(hom(D, D, m); check = false)
    push!(geneOAinOD, fD)
  end

  geneOBinOD = elem_type(OD)[]
  for f in gens(OB)
    m = block_diagonal_matrix(ZZMatrix[identity_matrix(ZZ, ngens(A)), matrix(f)])
    fD = OD(hom(D, D, m); check = false)
    push!(geneOBinOD, fD)
  end
  OAtoOD = hom(OA, OD, geneOAinOD; check = false)
  OBtoOD = hom(OB, OD, geneOBinOD; check = false)
  return D, AinD, BinD, OD, OAtoOD, OBtoOD
end

###############################################################################
#
#  Local tools
#
###############################################################################


# If fq is an isometry of the torsion module q, we compute the kernel of mu(fq)
# restricted to the p-elementary part of q (i.e. the submodule of q generated by
# all the elements of order p)
#
# This object is defined in Algorithm 2 of [BH23], the glue kernel we aim to use
# for gluing lattices in a p-admissible triples are actually submodules of these
# V's.
function _get_V(fq::TorQuadModuleMor, mu, p::IntegerUnion)
  q = domain(fq)
  V, _ = primary_part(q, p)
  _, Vinq = sub(q, elem_type(q)[q(lift(divexact(order(g), p)*g)) for g in gens(V) if !(order(g)==1)])
  fpV = restrict_endomorphism(fq, Vinq)
  fpV = evaluate(mu, fpV)
  V, _ = kernel(fpV)
  Vinq = hom(V, q, elem_type(q)[q(lift(a)) for a in gens(V)])
  @hassert :ZZLatWithIsom 1 is_injective(Vinq)
  return V, Vinq
end

# This is the rho function as defined in Definition 4.8 of BH23.
function _rho_functor(q::TorQuadModule, p, l::Union{Integer, fmpz}; quad::Bool = (p == 2))
  pq, pqtoq = primary_part(q, p)
  pq = rescale(pq, QQ(p)^(l-1))
  Nv = cover(pq)
  N = relations(pq)
  if quad && p == 2
    fr = _is_free(q, p, l)
  else
    fr = false
  end
  mqf = fr ? QQ(2) : QQ(1)
  if l == 0 
    Gl = N
    Gm = intersect(1//p*N, Nv)
    rholN = torsion_quadratic_module(Gl, p*Gm; modulus = QQ(1), modulus_qf = mqf)
  else
    k = l-1
    m = l+1
    Gk = intersect((1//(p^k))*N, Nv)
    Gl = intersect((1//(p^l))*N, Nv)
    Gm = intersect((1//(p^m))*N, Nv)
    B = Gk+p*Gm
    rholN = torsion_quadratic_module(Gl, B; modulus = QQ(1), modulus_qf = mqf)
  end
  return rholN
end

# A finite bilinear module over the 2-adic integers is even if all square are
# zeros.
function _is_even(T, p, l)
  B = gram_matrix_bilinear(_rho_functor(T, p, l; quad = false))
  return is_empty(B) || (all(is_zero, diagonal(B)) && all(is_integral, 2*B))
end

function _is_free(T, p, l)
  return _is_even(T, p, l-1) && _is_even(T, p, l+1)
end

###############################################################################
#
#  Overlattices
#
###############################################################################

# Compute the overlattice corresponding to the glue map gamma, in the ambient
# space of the covering lattice of the finite bilinear module D. As inputs,
# gamma must be an anti-isometry between HA and HB, both should embed in D and
# gamma commutes with the actions of fA and fB on HA and HB respectively.
#
# If `same_ambient = true`, then we consider all the problem in a same ambient
# quadratic space. In particular, the covering lattices of HA, HB and D are all
# in that same space.
#
# fA and fB here are considered as isometry of the relations lattice of HA and
# HB, respectively.
function _overlattice(gamma::TorQuadModuleMor,
                      HAinD::TorQuadModuleMor,
                      HBinD::TorQuadModuleMor,
                      fA::QQMatrix = identity_matrix(QQ, rank(relations(domain(HAinD)))),
                      fB::QQMatrix = identity_matrix(QQ, rank(relations(domain(HBinD))));
                      same_ambient::Bool = false)
  HA = domain(HAinD)
  HB = domain(HBinD)
  A = relations(HA)
  B = relations(HB)
  D = codomain(HAinD)
  if same_ambient
    _glue = Vector{QQFieldElem}[lift(g) + lift(gamma(g)) for g in gens(domain(gamma))]
    z = zero_matrix(QQ, 0, degree(A))
    glue = reduce(vcat, QQMatrix[matrix(QQ, 1, degree(A), g) for g in _glue]; init=z)
    glue = vcat(basis_matrix(A+B), glue)
    glue = FakeFmpqMat(glue)
    _B = hnf(glue)
    _B = QQ(1, denominator(glue))*change_base_ring(QQ, numerator(_B))
    C = lattice(ambient_space(A), _B[end-rank(A)-rank(B)+1:end, :])
    fC = block_diagonal_matrix([fA, fB])
    _B = solve_left(reduce(vcat, basis_matrix.([A,B])), basis_matrix(C))
    fC = _B*fC*inv(_B)
  else
    _glue = Vector{QQFieldElem}[lift(HAinD(a)) + lift(HBinD(gamma(a))) for a in gens(domain(gamma))]
    z = zero_matrix(QQ, 0, degree(cover(D)))
    glue = reduce(vcat, QQMatrix[matrix(QQ, 1, degree(cover(D)), g) for g in _glue]; init=z)
    glue = vcat(block_diagonal_matrix(basis_matrix.([A, B])), glue)
    glue = FakeFmpqMat(glue)
    _B = hnf(glue)
    _B = QQ(1, denominator(glue))*change_base_ring(QQ, numerator(_B))
    C = lattice(ambient_space(cover(D)), _B[end-rank(A)-rank(B)+1:end, :])
    fC = block_diagonal_matrix([fA, fB])
    _B = solve_left(block_diagonal_matrix(basis_matrix.([A, B])), basis_matrix(C))
    fC = _B*fC*inv(_B)
  end
  @hassert :ZZLatWithIsom 1 fC*gram_matrix(C)*transpose(fC) == gram_matrix(C)
  _, graph = sub(D, D.(_glue))
  return C, fC, graph
end

# Same as above where we glue along the trivial subgroups of HA and HB. In that
# particular case, HA and HB are the discriminant group of the lattices
# considered (so HA = L^{\vee}/L for some lattice integral lattice L, and same
# for HB), and D is the orthogonal direct sum of HA and HB in an appropriate
# quadratic space.
function _overlattice(HAinD::TorQuadModuleMor,
                      HBinD::TorQuadModuleMor,
                      fA::QQMatrix = identity_matrix(QQ, rank(relations(domain(HAinD)))),
                      fB::QQMatrix = identity_matrix(QQ, rank(relations(domain(HBinD))));
                      same_ambient::Bool = false)
  HA = domain(HAinD)
  HB = domain(HBinD)
  zA, _ = sub(HA, TorQuadModuleElem[])
  zB, _ = sub(HB, TorQuadModuleElem[])
  gamma = hom(zA, zB, zero_matrix(ZZ, 0, 0))
  return _overlattice(gamma, HAinD, HBinD, fA, fB; same_ambient)
end

##############################################################################
#
#  Orbits and stabilizers of discriminant subgroups
#
##############################################################################

function _on_subgroup_automorphic(T::TorQuadModule, g::AutomorphismGroupElem)
  q = domain(parent(g))
  gene = TorQuadModuleElem[g(q(lift(t))) for t in gens(T)]
  return sub(q, gene)[1]
end

# Compute stabilizer of a subgroup of a `TorQuadModule` under the action by
# automorphisms.
function stabilizer(O::AutomorphismGroup{TorQuadModule}, i::TorQuadModuleMor)
  @req domain(O) === codomain(i) "Incompatible arguments"
  q = domain(O)
  N, _ = sub(q, i.(gens(domain(i))))
  stab, _ = stabilizer(O, N, _on_subgroup_automorphic)
  return sub(O, O.([h.X for h in gens(stab)]))
end

function _subgroups_orbit_representatives_and_stabilizers(Vinq::TorQuadModuleMor,
                                                          O::AutomorphismGroup{TorQuadModule},
                                                          ord::IntegerUnion = -1,
                                                          f::Union{TorQuadModuleMor, AutomorphismGroupElem{TorQuadModule}} = id_hom(codomain(Vinq));
                                                          compute_stab::Bool = true)
  res = Tuple{TorQuadModuleMor, AutomorphismGroup{TorQuadModule}}[]

  V = domain(Vinq)
  q = codomain(Vinq)

  if !is_divisible_by(order(V), ord)
    return res
  end

  fV = f isa TorQuadModuleMor ? restrict_endomorphism(f, Vinq) : restrict_endomorphism(hom(f), Vinq)
  if ord == -1
    subs = collect(stable_submodules(V, TorQuadModuleMor[fV]))
  else
    subs = collect(submodules(V; order = ord))
    filter!(s -> is_invariant(fV, s[2]), subs)
  end

  subs = TorQuadModule[s[1] for s in subs]
  m = gset(O, _on_subgroup_automorphic, subs)
  orbs = orbits(m)
  for orb in orbs
    rep = representative(orb)
    if compute_stab
      stab, _ = stabilizer(O, rep, _on_subgroup_automorphic)
    else
      stab = O
    end
    _, rep = sub(q, TorQuadModuleElem[q(lift(g)) for g in gens(rep)])
    push!(res, (rep, stab))
  end
  return res
end

# The underlying abelian groups of H and V are elementary abelian p-groups, f is
# an automorphism of V fixing H, so in particular it acts on the quotient V/H
# whose abelian structure actually defines a finite dimensional Fp-vector space.
#
# This map return Qp := V/H as an Fp-vector space, the map which transforms V into a
# Fp-vector space Vp, the quotient map Vp \to Qp, and the restriction fQp of f
# to Qp
function _cokernel_as_Fp_vector_space(HinV::TorQuadModuleMor, p::IntegerUnion)
  H = domain(HinV)
  V = codomain(HinV)

  n = ngens(V)
  F = GF(p)
  Vp = VectorSpace(F, n)

  function _VtoVp(x::TorQuadModuleElem)
    v = data(x).coeff
    return Vp(vec(collect(v)))
  end

  function _VptoV(v::ModuleElem{FpFieldElem})
    x = lift.(v.v)
    return sum(TorQuadModuleElem[x[i]*V[i] for i in 1:ngens(V)])
  end

  VtoVp = Hecke.MapFromFunc(V, Vp, _VtoVp, _VptoV)
  subgene = elem_type(Vp)[VtoVp(HinV(a)) for a in gens(H)]
  Hp, _ = sub(Vp, subgene)
  Qp, VptoQp = quo(Vp, Hp)

  return Qp, VtoVp, VptoQp
end

# Almost duplicate of an existing function: we do not always want to compute
# stabilizers but just some orbit representatives
function _orbit_representatives(G::MatrixGroup{E}, k::Int, O::AutomorphismGroup{TorQuadModule}) where E <: FinFieldElem
  F = G.ring
  n = G.deg
  q = GAP.Obj(order(F))
  V = VectorSpace(F, n)
  orbs = GAP.Globals.Orbits(G.X, GAP.Globals.Subspaces(GAP.Globals.GF(q)^n, k))
  orbs = [GAP.Globals.BasisVectors(GAP.Globals.Basis(orb[1])) for orb in orbs]
  orbs = [[[F(x) for x in v] for v in bas] for bas in orbs]
  orbs = [sub(V, [V(v) for v in bas])[1] for bas in orbs]
  return [(orbs[i], O) for i in 1:length(orbs)]
end

# Given an abelian group injection V \to q where the group structure on V is
# abelian p-elementary, compute orbits and stabilizers of subgroups of V of
# order ord, which contains p^l*q and which are fixed under the action of f, under
# the action by automorphisms of q in G. 
#
# Note that V is fixed by the elements in G, f commutes with the elements in G
# and G is seen as a set of outer automorphisms (so two subgroups are in the
# same orbit if they are G-automorphic).
function _subgroups_orbit_representatives_and_stabilizers_elementary(Vinq::TorQuadModuleMor,
                                                                     G::AutomorphismGroup{TorQuadModule},
                                                                     ord::IntegerUnion,
                                                                     f::Union{TorQuadModuleMor, AutomorphismGroupElem{TorQuadModule}} = id_hom(codomain(Vinq)),
                                                                     l::IntegerUnion = -1;
                                                                     compute_stab::Bool = true)
  res = Tuple{TorQuadModuleMor, AutomorphismGroup{TorQuadModule}}[]

  V = domain(Vinq)

  # If V is trivial, then we ignore f and l, we just need to ensure that the
  # order wanted is also 1
  if order(V) == 1
    ord != 1 && (return res)
    push!(res, (Vinq, G))
    return res
  end

  q = codomain(Vinq)
  p = elementary_divisors(V)[1]
  pq, pqtoq = primary_part(q, p)
  l = l < 0 ? valuation(order(pq), p) : l
  g = valuation(ord, p)

  # some other trivial cases: if ord is 1, then l should be null (-1 by default)
  # Otherwise, if ord == order(V), since V is preserved by f and contained the
  # good subgroup of q, we just return V
  if ord == 1
    l < valuation(order(pq), p) && (return res)
    _, triv = sub(codomain(Vinq), TorQuadModuleElem[])
    push!(res, (triv, G))
    return res
  elseif ord == order(V)
    push!(res, (Vinq, G))
    return res
  end

  # In theory, V should contain H0 := p^l*pq where pq is the p-primary part of q
  all(a -> has_preimage(Vinq, (p^l)*pqtoq(a))[1], gens(pq)) || return res
  H0, H0inq = sub(q, TorQuadModuleElem[q(lift((p^l)*a)) for a in gens(pq)])
  @hassert :ZZLatWithIsom 1 is_invariant(f, H0inq)

  # H0 should be contained in the group we want. So either H0 is the only one
  # and we return it, or if order(H0) > ord, there are no subgroups as wanted
  if order(H0) >= ord
    order(H0) > ord && return res
    push!(res, (H0inq, G))
    return res
  end

  H0inV = hom(H0, V, TorQuadModuleElem[V(lift(a)) for a in gens(H0)])
  @hassert :ZZLatWithIsom 1 is_injective(H0inV)

  # Since V and H0 are elementary p-groups, they can be seen as finite
  # dimensional vector spaces over a finite field, and so is their quotient.
  # Moreover, subgroups of V of order ord and containing H0 are in bijections
  # with cosets in V/H0 of rank val_p(ord-order(H)) over the finite field F_p
  Qp, VtoVp, VptoQp = _cokernel_as_Fp_vector_space(H0inV, p)
  Vp = codomain(VtoVp)

  # Should never happen, but who knows...
  dim(Qp) == 0 && return res

  # We descend G to V for computing stabilizers later on
  GV, GtoGV = restrict_automorphism_group(G, Vinq)
  if compute_stab
    satV, j = kernel(GtoGV)
  end

  # Automorphisms in G preserved V and H0, since the construction of H0 is
  # natural. Therefore, the action of G descends to the quotient and we look for
  # invariants sub-vector spaces of given rank in the quotient (then lifting
  # generators and putting them with H0 will give us invariant subgroups as
  # wanted)
  act_GV = FpMatrix[change_base_ring(base_ring(Qp), matrix(gg)) for gg in gens(GV)]
  act_GV = FpMatrix[solve(VptoQp.matrix, g*VptoQp.matrix) for g in act_GV]
  MGp = matrix_group(base_ring(Qp), dim(Qp), act_GV)
  GVtoMGp = hom(GV, MGp, MGp.(act_GV); check = false)
  GtoMGp = compose(GtoGV, GVtoMGp)

  @hassert :ZZLatWithIsom g-ngens(snf(abelian_group(H0))[1]) < dim(Qp)
  
  F = base_ring(Qp)
  # K is H0 but seen a subvector space of Vp (which is V)
  k, K = kernel(VptoQp.matrix; side = :left)
  gene_H0p = elem_type(Vp)[Vp(vec(collect(K[i,:]))) for i in 1:k]
  if compute_stab
    orb_and_stab = orbit_representatives_and_stabilizers(MGp, g-k)
  else
    orb_and_stab = _orbit_representatives(MGp, g-k, G)
  end

  for (orb, stab) in orb_and_stab
    i = orb.map
    gene_orb = elem_type(Qp)[Qp(vec(collect(i(v).v))) for v in gens(domain(i))]
    gene_orb = elem_type(Vp)[preimage(VptoQp, v) for v in gene_orb]

    gene_orb = vcat(gene_orb, gene_H0p)
    gene_orb = elem_type(V)[preimage(VtoVp, Vp(v)) for v in gene_orb]
    gene_orb = elem_type(q)[image(Vinq, v) for v in gene_orb]
    orbq, orbqinq = sub(q, gene_orb)
    @hassert :ZZLatWithIsom 1 order(orbq) == ord
    # We keep only f-stable subspaces
    is_invariant(f, orbqinq) || continue

    if compute_stab
      stabq = AutomorphismGroupElem{TorQuadModule}[GtoMGp\(s) for s in gens(stab)]
      stabq, _ = sub(G, union!(stabq, gens(satV)))
      # Stabilizers should preserve the actual subspaces, by definition. so if we
      # have lifted since properly, this should hold..
      @hassert :ZZLatWithIsom 1 is_invariant(stabq, orbqinq)
    else
      stabq = G
    end
    push!(res, (orbqinq, stabq))
  end
  return res
end

function _classes_automorphic_subgroups(q::TorQuadModule,
                                        O::AutomorphismGroup{TorQuadModule},
                                        H::TorQuadModule;
                                        f::Union{TorQuadModuleMor, AutomorphismGroupElem{TorQuadModule}} = id_hom(domain(O)),
                                        compute_stab::Bool = true)
  res = Tuple{TorQuadModuleMor, AutomorphismGroup{TorQuadModule}}[]

  # Trivial case: we look for subgroups in a given primary part of q
  ok, p = is_primary_with_prime(H)
  if ok
    if is_elementary(H, p)
      _, Vinq = _get_V(id_hom(q), minimal_polynomial(identity_matrix(QQ, 1)), p)
      sors = _subgroups_orbit_representatives_and_stabilizers_elementary(Vinq, O, order(H), f; compute_stab)
    else
      _, Vinq = primary_part(q, p)
      sors = _subgroups_orbit_representatives_and_stabilizers(Vinq, O, order(H), f; compute_stab)
    end
    filter!(d -> is_isometric_with_isometry(domain(d[1]), H)[1], sors)
    return sors
  end

  # We inspect each primary part of q and look for orbit representatives and
  # stabilizers of automorphic subgroups which will be isometric to the given
  # primary part of H.
  #
  # First, we cut q as an orthogonal direct sum of its primary parts
  pds = sort(prime_divisors(order(q)))
  if compute_stab
    blocks = TorQuadModuleMor[primary_part(q, pds[1])[2]]
    ni = Int[ngens(domain(blocks[1]))]
    for i in 2:length(pds)
      _f = blocks[end]
      _, j = has_complement(_f)
      _T = domain(j)
      __f = primary_part(_T, pds[i])[2]
      push!(blocks, compose(__f, j))
      push!(ni, ngens(domain(__f)))
    end
    D, inj, proj = biproduct(domain.(blocks))
  else
    blocks = TorQuadModuleMor[primary_part(q, p)[2] for p in pds]
    D, inj, proj = biproduct(domain.(blocks))
  end
  _, phi = is_isometric_with_isometry(D, q)

  list_can = Vector{Tuple{TorQuadModuleMor, AutomorphismGroup{TorQuadModule}}}[]
  # We collect the possible subgroups for each primary part, with the stabilizer
  for i in 1:length(pds)
    p = pds[i]
    qpinq = blocks[i]
    qp = domain(qpinq)
    T, _ = primary_part(H, p)
    Oqp = restrict_automorphism_group(O, qpinq)[1]
    fqp = restrict_endomorphism(f, qpinq)
    if is_elementary(T, p)
      _, j = _get_V(id_hom(qp), minimal_polynomial(identity_matrix(QQ, 1)), p)
      sors = _subgroups_orbit_representatives_and_stabilizers_elementary(j, Oqp, order(T), fqp; compute_stab)
    else
      sors = _subgroups_orbit_representatives_and_stabilizers(id_hom(qp), Oqp, order(T), fqp; compute_stab)
    end
    filter!(d -> is_isometric_with_isometry(domain(d[1]), T)[1], sors)
    is_empty(sors) && return res
    push!(list_can, sors)
  end

  # We gather together: we do a big cartesian product, and we remember to
  # reconstruct the stabilizer. Since primary parts do not talk to each other,
  # we concatenate generators on an orthogonal direct sum of q into its primary
  # parts (as we do for computations of orthogonal groups in the non split
  # degenerate case)
  for lis in Hecke.cartesian_product_iterator(list_can; inplace = false)
    embs = TorQuadModuleMor[l[1] for l in lis]
    embs = TorQuadModuleMor[hom(domain(embs[i]), q, TorQuadModuleElem[blocks[i](domain(blocks[i])(lift(embs[i](a)))) for a in gens(domain(embs[i]))]) for i in 1:length(lis)]
    H2, _proj = direct_product(domain.(embs)...)
    _, H2inq = sub(q, TorQuadModuleElem[sum([embs[i](_proj[i](g)) for i in 1:length(lis)]) for g in gens(H2)])
    if compute_stab
      stabs = AutomorphismGroup{TorQuadModule}[l[2] for l in lis]
      genestab = ZZMatrix[]

      for i in 1:length(ni)
        nb = sum(ni[1:i-1])
        na = sum(ni[(i+1):end])
        Inb = identity_matrix(ZZ, nb)
        Ina = identity_matrix(ZZ, na)
        append!(genestab, ZZMatrix[block_diagonal_matrix([Inb, matrix(f), Ina]) for f in gens(stabs[i])])
      end

      genestab = TorQuadModuleMor[hom(D, D, g) for g in genestab]
      genestas = ZZMatrix[matrix(compose(compose(inv(phi), g), phi)) for g in genestab]
      stab = Oscar._orthogonal_group(q, unique(genestas); check = false)
      @hassert :ZZLatWithIsom is_invariant(stab, H2inq)
    else
      stab = O
    end
    push!(res, (H2inq, stab))
  end

  return res
end

#################################################################################
#
# Primitive embeddings for even lattices
#
#################################################################################

# We compute representatives of isomorphism classes of primitive extensions
# M \oplus N \to L, where we glue along HM and HN which are respectively
# isometric and anti-isometric to H.
#
# We follow the second definition of Nikulin, i.e. we classify up to the action
# of O(N). If `classification == :sublat`, we also classify them up to the
# action of O(M). If `classification == :first`, we return the first embedding
# computed.
function _isomorphism_classes_primitive_extensions(N::ZZLat,
                                                   M::ZZLat,
                                                   H::TorQuadModule,
                                                   classification::Symbol)
  @hassert :ZZLatWithIsom 1 classification in [:first, :emb, :sublat]
  results = Tuple{ZZLat, ZZLat, ZZLat}[]

  qN = discriminant_group(N)
  GN, _ = image_in_Oq(N)

  qM = discriminant_group(M)
  if classification == :emb
    GM = Oscar._orthogonal_group(qM, ZZMatrix[matrix(id_hom(qM))]; check = false)
  else
    GM, _ = image_in_Oq(M)
  end

  D, inj = direct_sum(qN, qM)
  qNinD, qMinD = inj
  OD = orthogonal_group(D)

  subsN = _classes_automorphic_subgroups(qN, GN, rescale(H, -1))
  @hassert :ZZLatWithIsom 1 !isempty(subsN)
  subsM = _classes_automorphic_subgroups(qM, GM, H)
  @hassert :ZZLatWithIsom 1 !isempty(subsM)

  for H1 in subsN, H2 in subsM
    ok, phi = is_anti_isometric_with_anti_isometry(domain(H1[1]), domain(H2[1]))
    @hassert :ZZLatWithIsom 1 ok

    HNinqN, stabN = H1
    HNinD = compose(HNinqN, qNinD)
    HN = domain(HNinqN)
    OHN = orthogonal_group(HN)

    HMinqM, stabM = H2
    HMinD = compose(HMinqM, qMinD)
    HM = domain(HMinqM)
    OHM = orthogonal_group(HM)

    actN = hom(stabN, OHN, elem_type(OHN)[OHN(restrict_automorphism(x, HNinqN)) for x in gens(stabN)])
    actM = hom(stabM, OHM, elem_type(OHM)[OHM(restrict_automorphism(x, HMinqM)) for x in gens(stabM)])
    imM, _ = image(actM)

    stabNphi = AutomorphismGroupElem{TorQuadModule}[OHM(compose(inv(phi), compose(hom(actN(g)), phi))) for g in gens(stabN)]
    stabNphi, _ = sub(OHM, stabNphi)

    if is_elementary_with_prime(HM)[1]
      iso = isomorphism(PermGroup, OHM)
    else
      iso = id_hom(OHM)
    end
    reps = double_cosets(codomain(iso), iso(stabNphi)[1], iso(imM)[1])
    @vprintln :ZZLatWithIsom 1 "$(length(reps)) isomorphism classe(s) of primitive extensions"

    for g in reps
      g = iso\(representative(g))
      phig = compose(phi, hom(g))
      L, _, _ = _overlattice(phig, HNinD, HMinD)
      N2 = lattice_in_same_ambient_space(L, hcat(basis_matrix(N), zero_matrix(QQ, rank(N), degree(M))))
      @hassert  :ZZLatWithIsom 1 genus(N) == genus(N2)
      M2 = lattice_in_same_ambient_space(L, hcat(zero_matrix(QQ, rank(M), degree(N)), basis_matrix(M)))
      @hassert :ZZLatWithIsom 1 genus(M) == genus(M2)
      push!(results, (L, M2, N2))
      @vprintln :ZZLatWithIsom 1 "Gluing done"
      classification == :first && return results
    end
  end
  return results
end

@doc raw"""
    primitive_embeddings(L::ZZLat, M::ZZLat; classification::Symbol = :sublat,
                                             check::Bool = true)
                                    -> Bool, Vector{Tuple{ZZLat, ZZLat, ZZLat}}

Given an even integer lattice `L`, which is unique in its genus, and an even
integer lattice `M`, return whether `M` embeds primitively in `L`.

The first input of the function is a boolean `T` stating whether or not `M`
embeds primitively in `L`. The second output `V` consists on triples
`(L', M', N')` where `L'` is isometric to `L`, `M'` is a primitive sublattice
of `L'` isometric to `M`, and `N'` is the orthogonal complement of `M'` in `L'`.

If `T == false`, then `V` will always be the empty list. If `T == true`, then
the content of `V` actually depends on the value of the symbol `classification`.
There are 4 possibilities:
  - `classification = :none`: `V` is the empty list;
  - `classification = :first`: `V` consists on the first primitive embedding found;
  - `classification = :sublat`: `V` consists on representatives for all isomorphism classes of primitive embeddings of `M` in `L`, up to the actions of $\bar{O}(M)$ and $O(q)$ where `q` is the discriminant group of `L`;
  - `classification = :emb`: `V` consists on representatives for all isomorphism classes of primitive sublattices of `L` isometric to `M` up to the action of $O(q)$ where `q` is the discriminant group of `L`.

If `check` is set to true, the function determines whether `L` is in fact unique
in its genus.

# Examples
We can use such primitive embeddings algorithm to classify embedding in unimodular
lattices

```jldoctest
julia> E8 = root_lattice(:E,8);

julia> A4 = root_lattice(:A,4);

julia> bool, pe = primitive_embeddings(E8, A4)
(true, Tuple{ZZLat, ZZLat, ZZLat}[(Integer lattice of rank 8 and degree 8, Integer lattice of rank 4 and degree 8, Integer lattice of rank 4 and degree 8)])

julia> pe
1-element Vector{Tuple{ZZLat, ZZLat, ZZLat}}:
 (Integer lattice of rank 8 and degree 8, Integer lattice of rank 4 and degree 8, Integer lattice of rank 4 and degree 8)

julia> genus(pe[1][2]) == genus(pe[1][3])
true
```
To be understood here that there exists a unique class of embedding of the root
lattice $A_4$ in the root lattice $E_8$, and the orthogonal primitive sublattice
is isometric to $A_4$.
"""
function primitive_embeddings(L::ZZLat, M::ZZLat; classification::Symbol = :sublat, check::Bool = true)
  @req is_even(L) && is_even(M) "At the moment, only primitive embeddings into even integer lattices are computable"
  if check
    @req length(genus_representatives(L)) == 1 "L must be unique in its genus"
  end
  return primitive_embeddings(genus(L), M; classification)
end

@doc raw"""
    primitive_embeddings(q::TorQuadModule, sign::Tuple{Int, Int}, M::ZZLat;
                                            classification::Symbol = :sublat)
                                        -> Bool, Vector{Tuple{ZZLat, ZZLat, ZZLat}}

Given a tuple `sign` of non-negative integers and a torsion quadratic module
`q` which define a genus symbol `G` for even integer lattices, return whether the
even integer lattice `M` embeds primitively in a lattice in `G`.

The first input of the function is a boolean `T` stating whether or not `M`
embeds primitively in a lattice in `G`. The second output `V` consists on
triples `(L', M', N')` where `L'` is a lattice in `G`, `M'` is a sublattice of
`L'` isometric to `M`, and `N'` is the orthogonal complement of `M'` in `L'`.

If `T == false`, then `V` will always be the empty list. If `T == true`, then
the content of `V` actually depends on the value of the symbol `classification`.
There are 4 possibilities:
  - `classification = :none`: `V` is the empty list;
  - `classification = :first`: `V` consists on the first primitive embedding found;
  - `classification = :sublat`: `V` consists on representatives for all isomorphism classes of primitive embeddings of `M` in lattices in `G`, up to the actions of $\bar{O}(M)$ and $O(q)$ where `q` is the discriminant group of a lattice in `G`;
  - `classification = :emb`: `V` consists on representatives for all isomorphism classes of primitive sublattices of lattices in `G` isometric to `M`, up to the action of $O(q)$ where `q` is the discriminant group of a lattice in `G`.

If the pair `(q, sign)` does not define a non-empty genus for integer lattices,
an error is thrown.
"""
function primitive_embeddings(q::TorQuadModule, sign::Tuple{Int, Int}, M::ZZLat; classification::Symbol = :sublat)
  @req is_even(M) "At the moment, only primitive embeddings into even integer lattices are computable"
  @req is_genus(q, sign) "Invariants define the empty genus"
  G = genus(q, sign)
  @req is_even(G) "At the moment, only primitive embeddings into even integer lattices are computable"
  return primitive_embeddings(G, M; classification)
end

@doc raw"""
    primitive_embeddings(G::ZZGenus, M::ZZLat; classification::Symbol = :sublat)
                                      -> Bool, Vector{Tuple{ZZLat, ZZLat, ZZLat}}

Given a genus symbol `G` for even integer lattices and an even integer
lattice `M`, return whether `M` embeds primitively in a lattice in `G`.

The first input of the function is a boolean `T` stating whether or not `M`
embeds primitively in a lattice in `G`. The second output `V` consists on
triples `(L', M', N')` where `L'` is a lattice in `G`, `M'` is a sublattice of
`L'` isometric to `M`, and `N'` is the orthogonal complement of `M'` in `L'`.

If `T == false`, then `V` will always be the empty list. If `T == true`, then
the content of `V` actually depends on the value of the symbol `classification`.
There are 4 possibilities:
  - `classification = :none`: `V` is the empty list;
  - `classification = :first`: `V` consists on the first primitive embedding found;
  - `classification = :sublat`: `V` consists on representatives for all isomorphism classes of primitive embeddings of `M` in lattices in `G`, up to the actions of $\bar{O}(M)$ and $O(q)$ where `q` is the discriminant of a lattice in `G`;
  - `classification = :emb`: `V` consists on representatives for all isomorphism classes of primitive sublattices of lattices in `G` isometric to `M`, up to the action of $O(q)$ where `q` is the discriminant group of a lattice in `G`.
"""
function primitive_embeddings(G::ZZGenus, M::ZZLat; classification::Symbol = :sublat)
  @req is_even(G) && is_even(M) "At the moment, only primitive embeddings into even integer lattices are computable"
  @req classification in [:none, :emb, :sublat, :first] "Wrong symbol for classification"
  posL, _, negL = signature_tuple(G)
  posM, _, negM = signature_tuple(M)
  @req (posL-posM >= 0 && negL-negM >= 0) "Impossible embedding"

  results = Tuple{ZZLat, ZZLat, ZZLat}[]
  if rank(M) == rank(G)
    genus(M) != G && return false, results
    push!(results, (M, M, orthogonal_submodule(M, M)))
    return true, results
  end

  @req rank(M) < rank(G) "The rank of M must be smaller or equal than the one of the lattices in G"

  qM = discriminant_group(M)
  OqM = orthogonal_group(qM)
  GM, _ = image_in_Oq(M)
  
  qL = discriminant_group(rescale(G, -1))
  GL = orthogonal_group(qL)

  D, inj = direct_sum(qM, qL)
  qMinD, qLinD = inj

  prG, pG = is_primary_with_prime(G)
  elG = is_elementary(G, pG)

  prM, pM = is_primary_with_prime(M)
  elM = is_elementary(M, pM)

  for k in divisors(gcd(order(qM), order(qL)))
    ok, ek, pk = is_prime_power_with_data(k)
    @vprintln :ZZLatWithIsom 1 "Glue order: $(k)"

    if elG || elM
      _, VLinqL = _get_V(id_hom(qL), minimal_polynomial(identity_matrix(QQ, 1)), max(pG, pM))
      subsL = _subgroups_orbit_representatives_and_stabilizers_elementary(VLinqL, GL, k; compute_stab = false)
    elseif ok && (ek == 1)
      _, VLinqL = _get_V(id_hom(qL), minimal_polynomial(identity_matrix(QQ, 1)), k)
      subsL = _subgroups_orbit_representatives_and_stabilizers_elementary(VLinqL, GL, k; compute_stab = false)
    else
      if prG || prM
        _, VLinqL = primary_part(qL, max(pG, pM))
      elseif ok
        _, VLinqL = primary_part(qL, pk)
      else
        VLinqL = id_hom(qL)
      end
      subsL = _subgroups_orbit_representatives_and_stabilizers(VLinqL, GL, k; compute_stab = false)
    end
    isempty(subsL) && continue

    if elG || elM
      VM, VMinqM = _get_V(id_hom(qM), minimal_polynomial(identity_matrix(QQ, 1)), max(pG, pM))
    elseif ok && (ek == 1)
      VM, VMinqM = _get_V(id_hom(qM), minimal_polynomial(identity_matrix(QQ, 1)), k)
    else
      if prG || prM
        VM, VMinqM = primary_part(qM, max(pG, pM))
      elseif ok
        VM, VMinqM = primary_part(qM, k)
      else
        VM = qM
        VMinqM = id_hom(qM)
      end
    end
    !is_divisible_by(order(VM), k) && continue

    itM = submodules(VM; order = k)
    st = iterate(itM)
    @hassert :ZZLatWithIsom !isnothing(st)

    @vprintln :ZZLatWithIsom 1 "$(length(subsL)) subgroup(s)"
    for H in subsL
      HL = H[1]
      _st = st
      HM, state = st
      while !isnothing(HM) && !is_anti_isometric_with_anti_isometry(HM[1], domain(HL))[1]
        _st = iterate(itM, state)
        if isnothing(_st)
          HM = _st
        else
          HM, state = _st
        end
      end

      isnothing(HM) && continue

      HM = compose(HM[2], VMinqM)
      @vprintln :ZZLatWithIsom 1 "Possible gluings"
      ok, phi = is_anti_isometric_with_anti_isometry(domain(HM), domain(HL))
      @hassert :ZZLatWithIsom 1 ok

      ext = Vector{QQFieldElem}[lift(qMinD(qM(lift((g))))) + lift(qLinD(qL(lift(phi(g))))) for g in gens(domain(HM))]
      ext, _ = sub(D, D.(ext))
      perp, j = orthogonal_submodule(D, ext)
      disc = torsion_quadratic_module(cover(perp), cover(ext); modulus = modulus_bilinear_form(perp),
                                                               modulus_qf = modulus_quadratic_form(perp))
      disc2 = rescale(disc, -1)
      !is_genus(disc2, (posL-posM, negL-negM))  && continue

      classification == :none && return true, results

      G2 = genus(disc2, (posL-posM, negL-negM))
      @vprintln :ZZLatWithIsom 1 "We can glue: $(G2)"
      Ns = representatives(G2)
      @vprintln :ZZLatWithIsom 1 "$(length(Ns)) possible orthogonal complement(s)"
      Ns = lll.(Ns)

      # We want to compute the subgroup of `qM` which is identified with a
      # subgroup if `disc`. For this, we need to project `disc` into `qM`.
      # Here `qL` is given as $L1/L2$ for some lattice $L2 \subset L1$. We
      # construct an overlattice `S` of $M\perp L2$ inside $Mv\perp L1$ with
      # respect to `phi`, and we have
      # $M\perp L2 \subset S\subset Sv\subset Mv\perpL1$
      #
      # To have the subgroup, we embed `qM` in `D`. This gives a group `TT`
      # which is of the form $T/M\perp L2$ for some overlattice
      # $M\perp L2 \subset T \subset Mv\perp L1$.
      #
      # The upshot is that both $(T\cap S)/(M\perp L2)$ and
      # $(T\cap Sv)/(M\perpL2) lies in the embedding of $qM \to D$. Their
      # quotient is isometric to the quotient of their preimage.
      # The quotient of their preimage is precisely the subgroup of `qM`
      # we look for.
      S, _, _ = _overlattice(phi, compose(HM, qMinD), compose(HL, qLinD))
      TT, _ = sub(D, qMinD.(gens(qM)))
      qM2 = torsion_quadratic_module(intersect(cover(TT), dual(S)),
                                     intersect(cover(TT), S);
                                     modulus = QQ(1),
                                     modulus_qf = QQ(2))

      for N in Ns
        temp = _isomorphism_classes_primitive_extensions(N, M, qM2, classification)
        if !is_empty(temp)
          classification == :first && return true, temp
          append!(results, temp)
        end
      end
    end
  end
  return (length(results) > 0), results
end

####################################################################################
#
# Admissible equivariant primitive extensions
#
####################################################################################

@doc raw"""
    admissible_equivariant_primitive_extensions(Afa::ZZLatWithIsom,
                                                Bfb::ZZLatWithIsom,
                                                Cfc::ZZLatWithIsom,
                                                p::Integer,
                                                q::Integer = p; check::Bool = true)
                                                     -> Vector{ZZLatWithIsom}

Given a triple of lattices with isometry `(A, fa)`, `(B, fb)` and `(C, fc)` and a
prime number `p`, such that `(A, B, C)` is `p`-admissible, return a set of
representatives of the double coset $G_B\backslash S/G_A$ where:

  - ``G_A`` and ``G_B`` are the respective images of the morphisms $O(A, fa) \to O(q_A, \bar{fa})$ and $O(B, fb) \to O(q_B, \bar{fb})$;
  - ``S`` is the set of all primitive extensions $A \perp B \subseteq C'$ with isometry $fc'$ where $p\cdot C' \subseteq A\perp B$ and such that the type of $(C', fc'^q)$ is equal to the type of `(C, fc)`.

If `check == true` the input triple is checked to a `p`-admissible triple of
integral lattices (with isometry) with `fA` and `fB` having relatively coprime
irreducible minimal polynomials and imposing that `A` and `B` are orthogonal
if `A`, `B` and `C` lie in the same ambient quadratic space.
"""
function admissible_equivariant_primitive_extensions(A::ZZLatWithIsom,
                                                     B::ZZLatWithIsom,
                                                     C::ZZLatWithIsom,
                                                     p::Integer,
                                                     q::Integer = p; check::Bool = true)
  # p and q can be equal, and they will be most of the time
  @req is_prime(p) && is_prime(q) "p and q must be a prime number" 

  # Requirements for [BH23]
  same_ambient = ambient_space(lattice(A)) === ambient_space(lattice(B)) === ambient_space(lattice(C))
  if check 
    @req all(L -> is_integral(L), [A, B, C]) "Underlying lattices must be integral"
    chiA = minimal_polynomial(A)
    chiB = minimal_polynomial(parent(chiA), isometry(B))
    @req gcd(chiA, chiB) == 1 "Minimal irreducible polynomials must be relatively coprime"
    @req is_admissible_triple(A, B, C, p) "Entries, in this order, do not define an admissible triple with respect to p"
    if same_ambient
      G = gram_matrix(ambient_space(C))
      @req iszero(basis_matrix(A)*G*transpose(basis_matrix(B))) "Lattices in same ambient space must be orthogonal"
    end
  end

  results = ZZLatWithIsom[]

  # this is the glue valuation: it is well-defined because the triple in input is admissible
  g = divexact(valuation(divexact(det(A)*det(B), det(C)), p), 2)

  qA, fqA = discriminant_group(A)
  qB, fqB = discriminant_group(B)
  qC = discriminant_group(lattice(C))
  GA, _ = image_centralizer_in_Oq(A)
  @hassert :ZZLatWithIsom 1 fqA in GA
  GB, _ = image_centralizer_in_Oq(B)
  @hassert :ZZLatWithIsom 1 fqB in GB

  # this is where we will perform the glueing
  if same_ambient
    D, qAinD, qBinD, OD, OqAinOD, OqBinOD = _sum_with_embeddings_orthogonal_groups(qA, qB)
  else
    D, qAinD, qBinD, OD, OqAinOD, OqBinOD = _direct_sum_with_embeddings_orthogonal_groups(qA, qB)
  end

  OqA = domain(OqAinOD)
  OqB = domain(OqBinOD)

  # if the glue valuation is zero, then we glue along the trivial group and we don't
  # have much more to do.
  if g == 0
    # Needed to compute the image of the stabilizer of the isometry we construct
    # (in the orthogonal group of the discriminant group of the new lattice).
    geneA = AutomorphismGroupElem{TorQuadModule}[OqAinOD(OqA(a.X)) for a in gens(GA)]
    geneB = AutomorphismGroupElem{TorQuadModule}[OqBinOD(OqB(b.X)) for b in gens(GB)]
    union!(geneA, geneB)

    # We compute the overlattice in this context
    C2, fC2, _ = _overlattice(qAinD, qBinD, isometry(A), isometry(B); same_ambient)
    C2fC2 = integer_lattice_with_isometry(C2, fC2; ambient_representation = false)

    # If not of the good type, we discard it
    !is_of_type(C2fC2^q, type(C)) && return results

    qC2 = discriminant_group(C2)
    OqC2 = orthogonal_group(qC2)
    phi2 = hom(qC2, D, TorQuadModuleElem[D(lift(x)) for x in gens(qC2)])
    @hassert :ZZLatWithIsom 1 is_isometry(phi2)

    # This is the new image of the stabilizer, just a direct product of
    # the previous ones
    GC2 = sub(OqC2, elem_type(OqC2)[OqC2(compose(phi2, compose(hom(g), inv(phi2))); check = false) for g in geneA])

    # This is mainly to check that we have not done anything inconsistent
    @hassert :ZZLatWithIsom 1 discriminant_group(C2fC2)[2] in GC2[1]
    set_attribute!(C2fC2, :image_centralizer_in_Oq, GC2)
    push!(results, C2fC2)
    return results
  end

  # these are GA|GB-invariant, fA|fB-stable, and should contain the kernels of any glue map.
  # VA and VB are submodules of the p-elementary parts of qA and qB
  # respectively.
  VA, VAinqA = _get_V(hom(fqA), minimal_polynomial(B), p)
  VB, VBinqB = _get_V(hom(fqB), minimal_polynomial(A), p)

  # since the glue kernels must have order p^g, in this condition, we have nothing
  if min(order(VA), order(VB)) < p^g
    return results
  end

  # scale of the dual: any glue kernel must contain the multiples of l of the respective
  # primary part of the discriminant groups
  l = valuation(level(genus(C)), p)

  # In the special case where rho_{l+1}(A) and rho_{l+1}(B) are free, then we
  # know that rho_l(C) is even if and only if any admissible gluing should
  # induce an isometry of finite quadratic form between rho_{l+1}(A) and
  # rho_{l+1}(B) (this works only when `p == 2`). In all the other cases, then
  # admissible gluings only induce isometries of finite bilinear modules between
  # rho_{l+1}(A) and rho_{l+1}(B).
  spec = (p == 2) && (_is_free(qA, p, l+1)) && (_is_free(qB, p, l+1)) && (_is_even(qC, p, l))

  # We look for the GA|GB-invariant and fA|fB-stable subgroups of VA|VB which respectively
  # contained lpqA|lpqB, where pqA and pqB are respectively the p-primary parts of qA and qB.
  # This is done by computing orbits and stabilisers of VA/lpqA (resp VB/lpqB)
  # seen as a F_p-vector space under the action of GA (resp. GB). Then we check which ones
  # are fA-stable (resp. fB-stable)
  subsA = _subgroups_orbit_representatives_and_stabilizers_elementary(VAinqA, GA, p^g, fqA, ZZ(l))
  subsB = _subgroups_orbit_representatives_and_stabilizers_elementary(VBinqB, GB, p^g, fqB, ZZ(l))

  # now, for each pair of anti-isometric potential kernels, we need to massage the gluing
  # computed to turn it into an admissible one. Then, we need to decide whether
  # such an admissible gluing can be made (fA,fB)-equivariant, up to conjugacy.
  #
  # Each pair for which we can find such nice gluing, we create the double coset
  # parametrising all such gluing up to certain conditions and we then compute the
  # corresponding overlattice and check whether it satisfies the type conditions.
  for H1 in subsA, H2 in subsB
    ok, phi = is_anti_isometric_with_anti_isometry(domain(H1[1]), domain(H2[1]))
    !ok && continue
    SAinqA, stabA = H1
    SA = domain(SAinqA)
    SAinD = compose(SAinqA, qAinD)
    OSA = orthogonal_group(SA)

    SBinqB, stabB = H2
    SB = domain(SBinqB)
    SBinD = compose(SBinqB, qBinD)
    OSB = orthogonal_group(SB)
    
    # we need a first admissible gluing. We know that such gluing exists because
    # we have an admissible triple as input and the glue kernels have been
    # chosen in such a way that their exist an admissible gluing between them.
    phi = _find_admissible_gluing(SAinqA, SBinqB, phi, l, spec)

    # we compute the image of the stabilizers in the respective OS* and we keep track
    # of the elements of the stabilizers acting trivially in the respective S*
    # (there are in the ker*).
    actA = hom(stabA, OSA, elem_type(OSA)[OSA(restrict_automorphism(x, SAinqA)) for x in gens(stabA)])
    imA, _ = image(actA)
    kerA = AutomorphismGroupElem{TorQuadModule}[OqAinOD(x) for x in gens(kernel(actA)[1])]
    push!(kerA, OqAinOD(one(OqA)))
    fSA = OSA(restrict_automorphism(fqA, SAinqA))

    actB = hom(stabB, OSB, elem_type(OSB)[OSB(restrict_automorphism(x, SBinqB)) for x in gens(stabB)])
    imB, _ = image(actB)
    kerB = AutomorphismGroupElem{TorQuadModule}[OqBinOD(x) for x in gens(kernel(actB)[1])]
    push!(kerB, OqBinOD(one(OqB)))
    fSB = OSB(restrict_automorphism(fqB, SBinqB))

    # We want all isometries of SB which preserves p^l*q_B and such that they
    # define isometries of rho_{l+1}(B). If `spec == true`, then rho_{l+1}(B) is
    # equipped with a quadratic form and we check isometries preserving it.
    # Otherwise, only isometries preserving the underlying bilinear product.
    OSBrB = _compute_double_stabilizer(SBinqB, l, spec)
    @hassert :ZZLatWithIsom 1 fSB in OSBrB   # Should always hold since the construction of rho_{l+1}(B) is natural in B
    fSB = OSBrB(fSB)

    # phi might not "send" the restriction of fA to this of fB, but at least phi*fA*phi^-1
    # should be conjugate to fB inside O(SB, rho_l(qB)) for the gluing.
    # If not, we try the next potential pair.
    fSAinOSB = OSB(compose(inv(phi), compose(hom(fSA), phi)))
    @hassert :ZZLatWithIsom 1 fSAinOSB in OSBrB  # Same as before, since phi is admissible, then the image of fSA should preserve rho_{l+1}(B)
    bool, g0 = is_conjugate_with_data(OSBrB, OSBrB(fSAinOSB), fSB)
    bool || continue

    # The new phi is "sending" the restriction of fA to this of fB
    # and it is still admissible. So we can glue SA and SB as wanted.
    phi = compose(phi, hom(OSB(g0)))
    @hassert :ZZLatWithIsom OSBrB(compose(inv(phi), compose(hom(fSA), phi))) == fSB

    # Now it is time to compute generators for O(SB, rho_l(qB), fB), and the induced
    # images of stabA|stabB for taking the double cosets next
    center, _ = centralizer(OSBrB, fSB)
    center, _ = sub(OSB, elem_type(OSB)[OSB(c) for c in gens(center)])
    stabSAphi, _ = sub(OSB, elem_type(OSB)[OSB(compose(inv(phi), compose(hom(g), phi))) for g in gens(imA)])
    stabSAphi, _ = intersect(center, stabSAphi)
    stabSB, _ = intersect(center, imB)

    iso = isomorphism(PermGroup, center)
    reps = double_cosets(codomain(iso), iso(stabSB)[1], iso(stabSAphi)[1])

    # We iterate over all double cosets. Each representative, define a new
    # classe of admissible gluing and so, for each such representative we compute the
    # corresponding overlattice along the gluing. If it has the wanted type, we compute
    # the image of the centralizer in OD from the stabA and stabB.
    for g in reps
      g = iso\(representative(g))
      phig = compose(phi, hom(g))
      @hassert :ZZLatWithIsom 1 is_anti_isometry(phig)

      # We compute the overlattice in this context, keeping track whether we
      # cork in a fixed ambient quadratic space
      C2, fC2, extinD = _overlattice(phig, SAinD, SBinD, isometry(A), isometry(B); same_ambient)
      C2fC2 = integer_lattice_with_isometry(C2, fC2; ambient_representation = false)

      # This is the type requirement: somehow, we want `(C2, fC2)` to be a "q-th root" of `(C, fC)`.
      !is_of_type(C2fC2^q, type(C)) && continue

      # By the theory of primitive extensions, the discriminant group qC2 of C2
      # is equal to H^{perp}/H where H is the graph of phig in D = qA + qB. We need
      # to treat both at the same time to compute the image of the centralizer
      # O(C2, fC2) in O(qC2, fqC2) using GA and GB.
      ext = domain(extinD)
      perp, j = orthogonal_submodule(D, ext)
      disc = torsion_quadratic_module(cover(perp), cover(ext); modulus = modulus_bilinear_form(perp),
                                                               modulus_qf = modulus_quadratic_form(perp))
      qC2 = discriminant_group(C2)
      OqC2 = orthogonal_group(qC2)
      phi2 = hom(qC2, disc, TorQuadModuleElem[disc(lift(x)) for x in gens(qC2)])
      @hassert :ZZLatWithIsom 1 is_isometry(phi2)              # In fact they are the same module so phi2, mathematically, is the identity.

      # So now this new integer lattice with isometry `(C2, fC2)` is a good
      # output. Just remain to compute GC2 in a smart way.
      geneOSA =  AutomorphismGroupElem{TorQuadModule}[OSA(compose(phig, compose(hom(g1), inv(phig)))) for g1 in unique(gens(imB))]
      im2_phi, _ = sub(OSA, geneOSA)
      im3, _, _ = intersect(imA, im2_phi)
      stab = AutomorphismGroupElem{TorQuadModule}[OqAinOD(actA\x) * OqBinOD(actB\(imB(compose(inv(phig), compose(hom(x), phig))))) for x in gens(im3)]
      union!(stab, kerA)
      union!(stab, kerB)
      stab = TorQuadModuleMor[restrict_automorphism(g, j) for g in stab]
      stab = TorQuadModuleMor[hom(disc, disc, [disc(lift(g(perp(lift(l))))) for l in gens(disc)]) for g in stab]
      stab = sub(OqC2, elem_type(OqC2)[OqC2(compose(phi2, compose(g, inv(phi2))); check = false) for g in stab])

      # If we have done good things, the action of fC2 on qC2 should centralize
      # itself...
      @hassert :ZZLatWithIsom 1 discriminant_group(C2fC2)[2] in stab[1]

      set_attribute!(C2fC2, :image_centralizer_in_Oq, stab)
      push!(results, C2fC2)
    end
  end
  return results
end

###############################################################################
#
#  Computation of O(H_B, rho_{l+1}(B))
#  ===================================
#  [BH23, Definition 4.16]
#
###############################################################################


# Action of isometries on the gram matrix of a finite bilinear form
function _on_modular_matrix(M::QQMatrix, g::AutomorphismGroupElem)
  q = domain(parent(g))
  R = Hecke.QmodnZ(QQ(1))
  m = matrix(inv(g))
  return map_entries(a -> lift(R(a)), m*M*transpose(m))
end

# Action of isometries on the gram matrix of a finite quadratic form
function _on_modular_matrix_quad(M::QQMatrix, g::AutomorphismGroupElem)
  q = domain(parent(g))
  R1 = Hecke.QmodnZ(QQ(1))
  R2 = Hecke.QmodnZ(QQ(2))
  m = matrix(inv(g))
  m1 = m*M*transpose(m)
  for i in 1:nrows(m1)
    for j in 1:ncols(m1)
      if i == j
        m1[i,j] = lift(R2(m1[i,j]))
      else
        m1[i,j] = lift(R1(m1[i,j]))
      end
    end
  end
  return m1
end

# We compute O(SB, rho_{l+1}(B)) where B has discriminant form qB. `spec` keep
# track whether rho_{l+1}(B) should be considered as a finite quadratic module
# or just a finite bilinear module (depends on the overlattice).
function _compute_double_stabilizer(SBinqB::TorQuadModuleMor, l::IntegerUnion, spec::Bool)
  SB = domain(SBinqB)
  qB = codomain(SBinqB)
  OSB = orthogonal_group(SB)
  p = elementary_divisors(SB)[1]
  rB = _rho_functor(qB, p, l+1)
  rBtoSB = hom(rB, SB, TorQuadModuleElem[SB(QQ(p^l)*lift(a)) for a in gens(rB)])
  HB, HBinSB = sub(SB, rBtoSB.(gens(rB)))
  OSBHB, _ = stabilizer(OSB, HBinSB)
  OHB, OSBHBtoOHB = restrict_automorphism_group(OSBHB, HBinSB)
  K, _ = kernel(OSBHBtoOHB)
  if spec
    OHBrB, _ = stabilizer(OHB, gram_matrix_quadratic(rB), _on_modular_matrix_quad)
  else
    OHBrB, _ = stabilizer(OHB, gram_matrix_bilinear(rB), _on_modular_matrix)
  end
  OSBrB, _ = sub(OSB, union!(OSB.(gens((OSBHBtoOHB\(OHBrB))[1])), gens(K)))
  return OSBrB
end


###############################################################################
#
#  Isometries between finite bilinear modules
#
###############################################################################

# Test whether an abelian group isomorphism respect the finite bilinear product
# of its domain (of type `TorQuadModule`).
function _is_isometry_bilinear(f::TorQuadModuleMor)
  !is_bijective(f) && return false
  for a in gens(domain(f))
    for b in gens(domain(f))
      if f(a)*f(b) != a*b
        return false
      end
    end
  end
  return true
end

# Test whether an abelian group isomorphism defines an anti isometry of the
# finite bilinear product defined on its domain (of type `TorQuadModule`)
function _is_anti_isometry_bilinear(f::TorQuadModuleMor)
  !is_bijective(f) && return false
  for a in gens(domain(f))
    for b in gens(domain(f))
      if f(a)*f(b) != -a*b
        return false
      end
    end
  end
  return true
end

# Compute an anti-isometry between the two finite bilinear modules r1 and r2.
function _anti_isometry_bilinear(r1::TorQuadModule, r2::TorQuadModule)
  @hassert :ZZLatWithIsom is_semi_regular(r1) === is_semi_regular(r2) === true
  hz = hom(r1, r2, zero_matrix(ZZ, ngens(r1), ngens(r2)))
  r2m = rescale(r2, -1)
  r2tor2m = hom(r2, r2m, identity_matrix(ZZ, ngens(r2)))
  r1N, r1tor1N = normal_form(r1)
  r2mN, r2mtor2mN = normal_form(r2m)
  gram_matrix_bilinear(r1N) == gram_matrix_bilinear(r2mN) || return false, hz
  T = hom(r1N, r2mN, identity_matrix(ZZ, ngens(r1N)))
  T = compose(r1tor1N, compose(T, compose(inv(r2mtor2mN), inv(r2tor2m))))
  @hassert :ZZLatWithIsom _is_anti_isometry_bilinear(T)
  return true, T
end

###############################################################################
#
#  Admissible gluings
#
###############################################################################

# If we are given a gluing between SA and SB, given that an admissible gluing
# exist, we massage phi until we turn it into an admissible gluing. There might
# be ways to improve such search, but I would expect that both loops iterating
# on OSB and OSBHB are terminating after few tries.
function _find_admissible_gluing(SAinqA::TorQuadModuleMor,
                                 SBinqB::TorQuadModuleMor,
                                 phi::TorQuadModuleMor,
                                 l::IntegerUnion,
                                 spec::Bool)
  SA = domain(SAinqA)
  SB = domain(SBinqB)
  p = elementary_divisors(SA)[1]
  qA = codomain(SAinqA)
  qB = codomain(SBinqB)
  pqA, pqAtoqA = primary_part(qA, p)
  pqB, pqBtoqB = primary_part(qB, p)
  rA = _rho_functor(qA, p, l+1)
  rB = _rho_functor(qB, p, l+1)
  rAtoSA = hom(rA, SA, TorQuadModuleElem[SA(QQ(p^l)*lift(a)) for a in gens(rA)])
  HA, _ = sub(SA, rAtoSA.(gens(rA)))
  rBtoSB = hom(rB, SB, TorQuadModuleElem[SB(QQ(p^l)*(lift(b))) for b in gens(rB)])
  HB, HBinSB = sub(SB, rBtoSB.(gens(rB)))
  if spec
    ok, phi_0 = is_anti_isometric_with_anti_isometry(rA, rB)
    @hassert :ZZLatWithIsom 1 ok
    @hassert :ZZLatWithIsom 1 is_anti_isometry(phi_0)
  else
    ok, phi_0 = _anti_isometry_bilinear(rA, rB)
    @hassert :ZZLatWithIsom 1 ok
    @hassert :ZZLatWithIsom 1 _is_anti_isometry_bilinear(phi_0)
  end
  phiHA, _ = sub(SB, TorQuadModuleElem[phi(SA(lift(a))) for a in gens(HA)])
  OSB = orthogonal_group(SB)
  g = one(OSB)
  for f in OSB
    g = f
    if cover(_on_subgroup_automorphic(phiHA, g)) == cover(HB)
      break
    end
  end
  phi_1 = compose(phi, hom(g))
  OSBHB, _ = stabilizer(OSB, HBinSB)
  g = one(OSBHB)
  for f in OSBHB
    g = f
    phif = compose(phi_1, hom(f))
    psi = hom(rA, rB, TorQuadModuleElem[rBtoSB\(phif(rAtoSA(a))) for a in gens(rA)])
    if matrix(psi) == matrix(phi_0)
      break
    end
  end
  phig = compose(phi_1, hom(g))
  return phig
end
